import logging
import functools
import json

import requests

from requests.compat import urljoin  # type: ignore
from requests.utils import prepend_scheme_if_needed

from ppa import appliance

from . import exceptions

logger = logging.getLogger(__name__)

BASE_URL = "/backend/v1"

PROXIES = {"https": appliance.api_proxy(), "http": appliance.api_proxy()}
CERT = appliance.cert()


def _form_url(api_type: str, endpoint: str) -> str:
    return urljoin(prepend_scheme_if_needed(f"{appliance.address()}", "https"), f"{BASE_URL}/{api_type}/{endpoint}")


def logged_request(request_func):
    @functools.wraps(request_func)
    def _logged_api_request(api_type, endpoint, **kwargs):
        url = _form_url(api_type, endpoint)
        logger.info(f"sending {request_func.__name__.upper()} request to {url} with arguments {kwargs}")
        response = request_func(api_type, endpoint, **kwargs)
        try:
            response_json = response.json()
        except json.decoder.JSONDecodeError:
            logger.error(response.text)
            raise exceptions.ApiRequestError(
                f"PPA API call failed. Status code: {response.status_code}, response: {response.text}."
            )
        if response.status_code == 400:
            raise exceptions.ApiRequestError(
                f"PPA API call failed. Status code: 400, "
                f"hint: {response_json.get('hint')}, "
                f"error message: {response_json.get('message')}."
            )
        if response.status_code == 403:
            logger.error(response_json)
            raise exceptions.PermissionDenied(
                "PPA API call failed. The user running the task does not have "
                f"the required {endpoint} permissions in the PPA appliance."
            )
        if response.status_code == 500:
            hint = response_json.get("hint", "no hint provided")
            details = response_json.get("details", "no details provided")
            message = response_json.get("message", "no message provided")
            logger.error(response_json)
            raise exceptions.ServerError(
                f"PPA API call failed due to a server error. " f"Message: {message}, Details: {details}, Hint: {hint}"
            )
        return response_json

    return _logged_api_request


@logged_request
def get(api_type: str, endpoint: str, **params):
    return requests.get(
        _form_url(api_type, endpoint),
        params=params,
        headers={"Accept": "application/json"},
        proxies=PROXIES,
        verify=CERT,
    )


@logged_request
def post(api_type: str, endpoint: str, **data):
    return requests.post(_form_url(api_type, endpoint), proxies=PROXIES, verify=CERT, json=data)
