import logging
import functools

from typing import Optional


from . import api, query_formatters, tasks
from .exceptions import raise_on_empty, reraise_with_context, PermissionDenied
from .records import create_records, Image, Task


logger = logging.getLogger(__name__)

_get = functools.partial(api.get, api_type="rest", endpoint="images")
_post = functools.partial(api.post, api_type="rest/rpc")


@create_records(Image)
def get_all():
    """ Get all Images from the PPA Appliance.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.

    :returns: A list of :ref:`images:Image Records`.

    Example:

        .. code-block:: python

            images.get_all()
    """
    return _get()


@create_records(Image)
def search(**search_params):
    """ Search for Images on the PPA Appliance.

    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.ParameterError: if any supplied search value has not been formatted by a
        function in :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`images:Image Records`.

    Example:

        .. code-block:: python

            images.search(name=query_formatters._eq("test.image"))
    """
    query_formatters.assert_formatted_values(**search_params)
    return _get(**search_params)


@create_records(Image)
def get_by_name(name: str):
    """ Get a named Image from the PPA Appliance.

    :param image_id: The name of the Image.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if the named image does not exist.

    :returns: A single :ref:`images:Image Records` object.

    Example:

        .. code-block:: python

            images.get_by_name("test.image")
    """
    with raise_on_empty(f"There are no images delegated to you with name '{name}'."):
        return _get(name=query_formatters._eq(name))[0]


@create_records(Image)
def get_by_id(image_id: str):
    """ Get an Image with a certain ID from the PPA Appliance.

    :param image_id: The ID of the Image.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if an image with the supplied ID does not exist.

    :returns: A single :ref:`images:Image Records` object.

    Example:

        .. code-block:: python

            images.get_by_id(7245)
    """
    with raise_on_empty(f"There are no images delegated to you with ID '{image_id}'."):
        return _get(id=query_formatters._eq(image_id))[0]


def start(*, image_id: str, payload: Optional[dict] = None) -> Task:
    """ Start a new task using an Image ID.

    :param image_id: The ID of the Image to start a task from.
    :param payload: An optional dictionary of payload information to send to the task.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions to
        start a task, or has no delegated image matching the supplied `image_id`.

    :returns: The started task as a single :ref:`tasks:Task Records` object.

    Example:

        .. code-block:: python

            new_task = images.start(image_id=7245, payload={"started_by": ppa.task.user()})
    """
    with reraise_with_context(
        PermissionDenied,
        f"PPA API call failed. Unable to start the task because supplied image ID {image_id} either "
        "does not exist, or has not been delegated to the user making the API call.",
    ):
        return tasks.get_by_uuid(_post(endpoint="start_task", image_id=image_id, payload=payload))
