from typing import Union


from typeguard import typechecked

from . import exceptions


class FormattedValue(str):
    def __new__(cls, string: Union[str, int], original_value: Union[str, int, tuple]):
        cls.original_value = original_value
        # If the following line is not ignored, you get this error:
        #
        # error: Too many arguments for "__new__" of "object"
        #
        # which looks like the str equivalent of this bug: https://github.com/python/typeshed/issues/2686
        return super().__new__(cls, string)  # type: ignore


def assert_formatted_values(**search_params: str):
    invalid_search_params = []
    for value in search_params.values():
        if not isinstance(value, FormattedValue):
            invalid_search_params.append(value)
    if invalid_search_params:
        raise exceptions.ParameterError(
            f"The following {len(invalid_search_params)} values have not been formatted by a "
            f"function in ppa_api.query_formatters: {invalid_search_params}."
        )


@typechecked
def _eq(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform an `equality` check against the supplied value when passed
    in to a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.

    .. note:: For `inequality` checks, use the :py:func:`query_formatters._neq` function.
    """
    return FormattedValue(f"eq.{value}", value)


@typechecked
def _ilike(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform a case insensitive `match` operation against the supplied value when
    passed in to a `search` function.

    This function supports the use of `*` as a wildcard.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.

    .. note:: For case sensitive queries, use the :py:func:`query_formatters._like` function.
    """
    return FormattedValue(f"ilike.{str(value).lower()}", value)


@typechecked
def _like(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform a case sensitive `match` operation against the supplied value when
    passed in to a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.

    .. note:: For case insensitive queries, use the :py:func:`query_formatters._ilike` function.
    """
    return FormattedValue(f"like.{value}", value)


@typechecked
def _neq(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform an `inequality` check against the supplied value when passed in
    to a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.

    .. note:: For `equality` checks, use the :py:func:`query_formatters._eq` function.
    """
    return FormattedValue(f"neq.{value}", value)


@typechecked
def _lt(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform a `less than` check against the supplied value when passed in to
    a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.
    """
    return FormattedValue(f"lt.{value}", value)


@typechecked
def _lte(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform a `less than or equal to` check against the supplied value when
    passed in to a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.
    """
    return FormattedValue(f"lte.{value}", value)


@typechecked
def _gt(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform a `greater than` check against the supplied value when passed in
    to a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.
    """
    return FormattedValue(f"gt.{value}", value)


@typechecked
def _gte(value: Union[str, int]) -> FormattedValue:
    """
    Takes the supplied value and returns an API-ready query string. The query string
    will perform a `greater than or equal to` check against the supplied value when
    passed in to a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.
    """
    return FormattedValue(f"gte.{value}", value)


@typechecked
def _in(values: tuple) -> FormattedValue:
    """
    Takes the supplied tuple of values and returns an API-ready query string. The query string
    will perform an `in` operation against the supplied value when passed in to a `search` function.

    :param value: The value to generate the query string from.
    :returns: A FormattedValue object ready to pass in to a search function.
    """
    in_value = ",".join(f'"{i}"' for i in values)
    return FormattedValue(f"in.({in_value})", values)


def _not(value: FormattedValue) -> FormattedValue:
    """
    Takes an API-ready query string, and returns a new one that negates the supplied condition. Use this function
    to negate queries built by the other formatters in this module.

    :param value: The API-ready query string to negate.
    :returns: A FormattedValue object ready to pass in to a search function.

    .. important:: This function must be provided a value returned by another query formatter in this module.
    """
    if not isinstance(value, FormattedValue):
        raise exceptions.ParameterError(
            "All values supplied to the _not function must have been generated by another formatter."
        )
    return FormattedValue(f"not.{value}", value)
