import logging
import functools
import datetime

from typing import Optional, List, Union
from distutils.util import strtobool

import dateutil.parser

from datacaster.classes import CastDataClass


from . import exceptions

logger = logging.getLogger(__name__)


def create_records(record_type):
    @functools.singledispatch
    def _unpack(response_data):
        logger.error(f"unhandled response type: {response_data}")
        raise exceptions.ApiRequestError(
            "Did not receive either a single record or a list of records from the "
            "PPA API. Please view the logs for more information."
        )

    @_unpack.register(list)
    def _list_unpack(response_data):
        return [record_type(**record) for record in response_data]

    @_unpack.register(dict)
    def _dict_unpack(response_data):
        return record_type(**response_data)

    def _wrapped_api_function(func):
        @functools.wraps(func)
        def _execute_api_function(*args, **kwargs):
            return _unpack(func(*args, **kwargs))

        return _execute_api_function

    return _wrapped_api_function


def _process_datetime(value: str) -> datetime.datetime:
    # We don't use the datetime iso parser because sometimes postgrest
    # doesn't left 0-pad fractions of a second properly, which leads
    # to a ValueError. The dateutil isoparser handles this for us.
    return dateutil.parser.isoparse(value)


def _optional_datetime(value: Union[str, None]) -> Union[datetime.datetime, None]:
    return _process_datetime(value) if value else None


def _str_to_bool(value: str) -> bool:
    # Convert 't' or 'f' to True or False. Uses bool() on the return value of
    # strtobool, because strtobool actually returns an int.
    return bool(strtobool(value))


class Image(CastDataClass):
    # Images are the task cards you see in the UI. A Task record is created when an Image is started.
    __class_config__ = {
        "cast_functions": {
            "types": {
                datetime.datetime: _process_datetime,
                Optional[datetime.datetime]: _optional_datetime,
                bool: _str_to_bool,
            }
        },
    }

    id: int
    name: str
    description: str
    cost: Optional[str]
    hash: str
    author: str
    tags: List[str]
    updated_at: Optional[datetime.datetime]
    icon: Optional[str]
    os: str
    agent: str
    available: bool
    groups: int
    owner: Optional[str]  # The welcome image has no owner (it is baked in).


class Task(CastDataClass):
    # Tasks are executed instances of an Image and can be in various states, including 'running'.
    __class_config__ = {
        "cast_functions": {
            "types": {
                datetime.datetime: _process_datetime,
                Optional[datetime.datetime]: _optional_datetime,
                bool: _str_to_bool,
            }
        },
    }

    id: int
    image: str
    image_id: str
    uuid: str
    hash: str
    author: str
    username: str
    is_owner: bool  # Indicates whether the task was started by the API-call user.
    exit_code: Optional[int]  # There is no exit code if the task is running.
    defunct: bool
    started_at: Optional[datetime.datetime]
    stopped_at: Optional[datetime.datetime]
    cancelled_at: Optional[datetime.datetime]
    cancelled_by: Optional[str]
    timed_out: bool
    state: str
    agent_id: int
    timeout: int
    source: str


class Agent(CastDataClass):
    id: int
    name: str
    host: Optional[str]  # The PPA appliance is an agent, and has a None host.
    state: str


class User(CastDataClass):
    __class_config__ = {
        "cast_functions": {
            "types": {
                datetime.datetime: _process_datetime,
                Optional[datetime.datetime]: _optional_datetime,
                bool: _str_to_bool,
            }
        },
    }

    id: int
    name: str
    username: str
    email: str
    active: bool
    permissions: List[str]
    authenticated_at: Optional[datetime.datetime]
