import logging
import functools

from . import api, query_formatters
from .records import create_records, Task
from .exceptions import raise_on_empty

logger = logging.getLogger(__name__)

_get = functools.partial(api.get, api_type="rest", endpoint="tasks")


@create_records(Task)
def get_all():
    """ Get all Tasks from the PPA Appliance.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.

    :returns: A list of :ref:`tasks:Task Records`.

    Example:

        .. code-block:: python

            tasks.get_all()
    """
    return _get()


@create_records(Task)
def get_by_uuid(uuid: str):
    """ Get a Task from the PPA Appliance using its UUID.

    :param uuid: The UUID of the task.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if no task with the supplied UUID exists.

    :returns: A single :ref:`tasks:Task Records` object.

    Example:

        .. code-block:: python

            tasks.get_by_uuid("e8952f03-d7c4-4e2c-ba73-85c06f2a3e01")
    """
    with raise_on_empty(f"There are no tasks visible to you with UUID '{uuid}'."):
        return _get(uuid=query_formatters._eq(uuid))[0]


@create_records(Task)
def search(**search_params):
    """ Search for Tasks on the PPA Appliance.

    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.ParameterError: if any supplied search value has not been formatted by a
        function in :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`tasks:Task Records`.

    Example:

        .. code-block:: python

            tasks.search(username=query_formatters._eq("john.smith"))
    """
    query_formatters.assert_formatted_values(**search_params)
    return _get(**search_params)


@create_records(Task)
def started_by_user(username: str):
    """ Get a list of Task that were started by the supplied user.

    :param username: The name of the user who started the tasks.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if no task with the supplied UUID exists.

    :returns: A list of :ref:`tasks:Task Records`.

    Example:

        .. code-block:: python

            tasks.started_by_user("john.smith")
    """
    return _get(username=query_formatters._eq(username))
