import logging
import functools

from typing import Union

from . import api, query_formatters
from .records import create_records, User
from .exceptions import raise_on_empty, ServerError, NoSuchObject

logger = logging.getLogger(__name__)

_get = functools.partial(api.get, api_type="rest", endpoint="users")
_post = functools.partial(api.post, api_type="rest/rpc")


@create_records(User)
def get_all():
    """ Get all Users from the PPA Appliance.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.

    :returns: A list of :ref:`users:User Records`.

    Example:

        .. code-block:: python

            users.get_all()
    """
    return _get()


@create_records(User)
def search(**search_params):
    """ Search for Users on the PPA Appliance.

    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.ParameterError: if any supplied search value has not been formatted by a
        function in :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`users:User Records`.

    Example:

        .. code-block:: python

            users.search(username=query_formatters._eq("test.user"))
    """
    query_formatters.assert_formatted_values(**search_params)
    return _get(**search_params)


@create_records(User)
def get_by_name(name: str):
    """ Get a named User from the PPA Appliance.

    :param name: The name of the User.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if the named user does not exist.

    :returns: A single :ref:`users:User Records` object.

    Example:

        .. code-block:: python

            users.get_by_name("test.user")
    """
    with raise_on_empty(f"No user found with name '{name}'."):
        return _get(name=query_formatters._eq(name))[0]


@create_records(User)
def get_by_id(user_id: Union[str, int]):
    """ Get a User with a certain ID from the PPA Appliance.

    :param image_id: The ID of the User.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if a user with the supplied ID does not exist.

    :returns: A single :ref:`users:User Records` object.

    Example:

        .. code-block:: python

            users.get_by_id(123)
    """
    with raise_on_empty(f"No user found with ID '{user_id}'."):
        return _get(id=query_formatters._eq(user_id))[0]


def activate_by_id(user_id: Union[str, int]) -> None:
    """ Activate a user by ID in the PPA appliance.

    :param user_id: The ID of the User.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if a user with the supplied ID does not exist.

    :returns: None.

    Example:

        .. code-block:: python

            users.activate_by_id(123)
    """
    try:
        _post(endpoint="set_active", user_id=user_id, active=True)
    except ServerError as e:
        if "query returned no rows" in str(e):
            raise NoSuchObject(f"No user found with ID '{user_id}'.")
        raise e


def deactivate_by_id(user_id: Union[str, int]) -> None:
    """ Deactivate a user by ID in the PPA appliance.

    :param user_id: The ID of the User.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if a user with the supplied ID does not exist.

    :returns: None.

    Example:

        .. code-block:: python

            users.deactivate_by_id(123)
    """
    try:
        _post(endpoint="set_active", user_id=user_id, active=False)
    except ServerError as e:
        if "query returned no rows" in str(e):
            raise NoSuchObject(f"No user found with ID '{user_id}'.")
        raise e


def activate_by_name(name: str) -> None:
    """ Activate a named user in the PPA appliance.

    :param name: The name of the User.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if the named user does not exist.

    :returns: None.

    Example:

        .. code-block:: python

            users.activate_by_name("test.user")
    """
    _post(endpoint="set_active", user_id=get_by_name(name).id, active=True)


def deactivate_by_name(name: str) -> None:
    """ Deactivate a named user in the PPA appliance.

    :param name: The name of the User.

    :raises exceptions.PermissionDenied: if the user who triggered the task does not have the required permissions.
    :raises exceptions.NoSuchObject: if the named user does not exist.

    :returns: None.

    Example:

        .. code-block:: python

            users.deactivate_by_name("test.user")
    """
    _post(endpoint="set_active", user_id=get_by_name(name).id, active=False)
